@@ -11,23 +11,15 @@ use gettextrs::{LocaleCategory, gettext};
 use glib::ExitCode;
 use gtk::{gio, glib};
 
-use self::application::App;
+use self::application::ImpressionApp;
 use self::config::{GETTEXT_PACKAGE, LOCALEDIR, RESOURCES_FILE};
 
-#[macro_export]
-macro_rules! spawn {
-    ($future:expr) => {
-        let ctx = glib::MainContext::default();
-        ctx.spawn_local($future);
-    };
-}
-
 fn get_size_string(bytes_size: u64) -> String {
-    let mbs = bytes_size / 1024 / 1024;
-    match mbs {
-        0..=1023 => format!("{mbs}MB"),
+    let mebi_bytes = bytes_size / 1024 / 1024;
+    match mebi_bytes {
+        0..=1023 => format!("{mebi_bytes}MiB"),
         _ => {
-            format!("{:.2}GB", (mbs as f64) / 1024.0)
+            format!("{:.2}GiB", (mebi_bytes as f64) / 1024.0)
         }
     }
 }
@@ -37,7 +29,6 @@ fn runtime() -> &'static tokio::runtime::Runtime {
     RUNTIME.get_or_init(|| {
         tokio::runtime::Builder::new_multi_thread()
             .enable_all()
-            .thread_stack_size(12 * 1024 * 1024)
             .build()
             .expect("Setting up tokio runtime needs to succeed.")
     })
@@ -45,7 +36,7 @@ fn runtime() -> &'static tokio::runtime::Runtime {
 
 fn main() -> ExitCode {
     // Initialize logger
-    pretty_env_logger::init();
+    tracing_subscriber::fmt::init();
 
     // Prepare i18n
     gettextrs::setlocale(LocaleCategory::LcAll, "");
@@ -57,8 +48,6 @@ fn main() -> ExitCode {
     let res = gio::Resource::load(RESOURCES_FILE).expect("Could not load gresource file");
     gio::resources_register(&res);
 
-    // let _ = collect_online_distros();
-
-    let app = App::new();
+    let app = ImpressionApp::default();
     app.run()
 }
